from .item import Item

class Genre:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getName(self) -> str:
        return self.__item.getByName("value")

    def getAlbumCount(self) -> str:
        return self.__item.getByName("albumCount")

    def getSongCount(self) -> str:
        return self.__item.getByName("songCount")


from .item import Item
from .genre import Genre

class Genres:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getGenres(self) -> list[Genre]:
        return list(map(
            lambda x : Genre(x), 
            self.__item.getList(["genres", "genre"])))


from .item import Item
from .internet_radio_station import InternetRadioStation

class InternetRadioStations:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getStations(self) -> list[InternetRadioStation]:
        return list(map(
            lambda x : InternetRadioStation(x), 
            self.__item.getList(["internetRadioStations", "internetRadioStation"])))


from .item import Item
from .artist_list_item import ArtistListItem

class ArtistsInitial:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getName(self) -> str:
        return self.__item.getByName("name")

    def getArtistListItems(self) -> list[ArtistListItem]:
        return list(map(
            lambda x : ArtistListItem(x), 
            self.__item.getList(["artist"])))


from .item import Item

#from typing import Dict, Generic, TypeVar
from typing import Generic, TypeVar

T = TypeVar("T")

class Response(Generic[T]):

    def __init__(self, data : dict, t : T):
        self.__item : Item = Item(data)
        self.__t = t

    def getObj(self) -> T:
        return self.__t

    def getStatus(self) -> str:
        return self.__item.getByName("status")

    def getVersion(self) -> str:
        return self.__item.getByName("version")
    
    def isOk(self) -> bool:
        return self.getStatus() == "ok"


from .item import Item
from .playlist_entry import PlaylistEntry


class Playlist:

    __dict_name: str = "playlist"

    def __init__(self, data: dict):
        self.__item: Item = Item(data)
        select_item: self.__item
        if self.__item.isResponse() and Playlist.__dict_name in data:
            self.__is_response = True
            self.__response = self.__item
            select_item = Item(data[Playlist.__dict_name])
        else:
            self.__is_response = False
            self.__response = None
            select_item = self.__item
        self.__select_item: Item = select_item

    def getItem(self): return self.__select_item

    def getId(self) -> str:
        return self.__select_item.getByName("id")

    def getName(self) -> str:
        return self.__select_item.getByName("name")

    def getCreated(self) -> str:
        return self.__select_item.getByName("created")

    def getChanged(self) -> str:
        return self.__select_item.getByName("changed")

    def getCoverArt(self) -> str:
        return self.__select_item.getByName("coverArt")

    def getDuration(self) -> str:
        return self.__select_item.getByName("duration")

    def getOwner(self) -> str:
        return self.__select_item.getByName("owner")

    def getPublic(self) -> str:
        return self.__select_item.getByName("public")

    def getSongCount(self) -> str:
        return self.__select_item.getByName("songCount")

    def getEntries(self) -> list[PlaylistEntry]:
        result: list[PlaylistEntry] = list(map(
            lambda x: PlaylistEntry(x),
            self.__select_item.getList(["entry"])))
        return result



from .item import Item
from .playlist import Playlist

class Playlists:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getPlaylists(self) -> list[Playlist]:
        return list(map(
            lambda x : Playlist(x), 
            self.__item.getList(["playlists", "playlist"])))


from .item import Item
from .artist import Artist
from .album import Album
from .song import Song


class Starred:

    __first_entry_name: str = "starred2"

    def __init__(self, data: dict):
        self.__item: Item = Item(data)

    def getItem(self):
        return self.__item

    def getArtists(self) -> list[Artist]:
        return list(map(
            Artist,
            self.__item.getList([Starred.__first_entry_name, "artist"])))

    def getAlbums(self) -> list[Album]:
        return list(map(
            Album,
            self.__item.getList([Starred.__first_entry_name, "album"])))

    def getSongs(self) -> list[Song]:
        return list(map(
            Song,
            self.__item.getList([Starred.__first_entry_name, "song"])))



from .item import Item
from .multi_value import get_multi

class Song:

    __dict_name : str = "song"

    def __init__(self, data : dict):
        self.__item : Item = Item(data)
        select_item : self.__item
        if self.__item.isResponse() and Song.__dict_name in data:
            self.__is_response = True
            self.__response = self.__item
            select_item = Item(data[Song.__dict_name])
        else:
            self.__is_response = False
            self.__response = None
            select_item = self.__item
        self.__select_item : Item = select_item

    def getItem(self): return self.__select_item

    def getId(self) -> str:
        return self.__select_item.getId()

    def getTitle(self) -> str:
        return self.__select_item.getByName("title")

    def getArtist(self) -> str:
        return self.__select_item.getByName("artist")

    def getArtistId(self) -> str:
        return self.__select_item.getByName("artistId")

    def getAlbum(self) -> str:
        return self.__select_item.getByName("album")

    def getAlbumId(self) -> str:
        return self.__select_item.getByName("albumId")

    def getYear(self) -> str:
        return self.__select_item.getByName("year")

    def getDiscNumber(self) -> str:
        return self.__select_item.getByName("discNumber")

    def getTrack(self) -> str:
        return self.__select_item.getByName("track")

    def getCoverArt(self) -> str:
        return self.__select_item.getByName("coverArt")

    def getContentType(self) -> str:     
        return self.__select_item.getByName("contentType")

    def getBitRate(self) -> int:
        return self.__select_item.getByName("bitRate")

    def getSuffix(self) -> str:
        return self.__select_item.getByName("suffix")

    def getDuration(self) -> int:
        return self.__select_item.getByName("duration")

    def getParent(self) -> str:        
        return self.__select_item.getByName("parent")

    def getGenre(self) -> str:        
        genre_list : list[str] = self.__get_genres()
        return genre_list[0] if genre_list and len(genre_list) > 0 else None

    def getGenres(self) -> list[str]:
        return self.__get_genres()

    def __get_genres(self) -> list[str]:
        return get_multi(self.__select_item, "genre", "genres", "name")

    def getPath(self) -> str:
        return self.__select_item.getByName("path")
        
    def getCreated(self) -> str:        
        return self.__select_item.getByName("created")

    def getStarred(self) -> str:        
        return self.__select_item.getByName("starred")



from enum import Enum

class ListType(Enum):
    RANDOM = 0, "random"
    NEWEST = 1, "newest"
    HIGHEST = 2, "highest"
    FREQUENT = 3, "frequent"
    RECENT = 4, "recent"
    STARRED = 5, "starred"
    ALPHABETICAL_BY_NAME = 6, "alphabeticalByName"
    ALPHABETICAL_BY_ARTIST = 7, "alphabeticalByArtist"
    BY_YEAR = 8, "byYear"
    BY_GENRE = 9, "byGenre"

    def __init__(self, num, argValue : str):
        self.num = num
        self.argValue = argValue

    def getArgValue(self):
        return self.argValue


from .item import Item
from .artist import Artist
from .album import Album
from .song import Song

class SearchResult:

    __main_element : str = "searchResult3"

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getArtists(self) -> list[Artist]:
        return list(map(
            lambda x : Artist(x), 
            self.__item.getList([SearchResult.__main_element, "artist"])))

    def getAlbums(self) -> list[Album]:
        return list(map(
            lambda x : Album(x), 
            self.__item.getList([SearchResult.__main_element, "album"])))

    def getSongs(self) -> list[Song]:
        return list(map(
            lambda x : Song(x), 
            self.__item.getList([SearchResult.__main_element, "song"])))



from .item import Item
from .song import Song

class RandomSongs:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getSongs(self) -> list[Song]:
        return list(map(
            lambda x : Song(x), 
            self.__item.getList(["randomSongs", "song"])))


class ConfigurationMeta(type):
    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (hasattr(subclass, 'getBaseUrl') and 
                callable(subclass.getBaseUrl) and
                hasattr(subclass, 'getPort') and
                callable(subclass.getPort) and
                hasattr(subclass, 'getUserName') and
                callable(subclass.getUserName) and
                hasattr(subclass, 'getPassword') and
                callable(subclass.getPassword) and
                hasattr(subclass, 'getLegacyAuth') and
                callable(subclass.getLegacyAuth) and
                hasattr(subclass, 'getApiVersion') and
                callable(subclass.getApiVersion) and
                hasattr(subclass, 'getAppName') and
                callable(subclass.getAppName))

class ConfigurationInterface:
    def getBaseUrl(self) -> str: pass
    def getPort(self) -> int: pass
    def getUserName(self) -> str: pass
    def getPassword(self) -> str: pass
    def getApiVersion(self) -> str: pass
    def getAppName(self) -> str: pass
    def getLegacyAuth(self) -> bool: pass

class Configuration(metaclass = ConfigurationMeta):
    pass



from .item import Item

def get_multi(
        data : Item, 
        single_value_name : str, 
        multi_value_name : str, 
        dict_name : str = "name") -> list[str]:
    result : list[str] = list()
    single_value : str = data.getByName(single_value_name)
    if single_value: result.append(single_value)
    multi_value : list[dict[str, any]] = data.getListByName(multi_value_name, list())
    curr : dict[str, any]
    for curr in multi_value:
        v : str = curr[dict_name] if dict_name in curr else None
        if v and not v in result: result.append(v)
    return result


from .item import Item

class InternetRadioStation:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getId(self) -> str:
        return self.__item.getByName("id")

    def getName(self) -> str:
        return self.__item.getByName("name")

    def getStreamUrl(self) -> str:
        return self.__item.getByName("streamUrl")

    def getHomePageUrl(self) -> str:
        return self.__item.getByName("homePageUrl")


class Item:

    def __init__(self, data : dict):
        self.__data = data

    def getData(self) -> dict[str, any]:
        return self.__data

    def isResponse(self) -> bool:
        return (
            "status" in self.getData() and
            "version" in self.getData())

    def hasName(self, name : str) -> bool:
        return name in self.getData()

    def getByName(self, name : str, defaultValue = None) -> any:
        return (self.getData()[name]
            if name in self.getData() 
            else defaultValue)

    def getListByName(self, name : str, defaultValue = []) -> list[dict[str, any]]:
        return (self.getData()[name]
            if name in self.getData() 
            else defaultValue)
    
    def getId(self) -> str:
        return self.getByName("id")

    def getList(self, path : list[str], allowEmpty = True) -> list:
        result = self.getData()
        intermediate : str = ""
        for current_path in path:
            if not type(result) == dict:
                raise Exception(
                    "Intermediate path [{}] does not match a dictionary"
                        .format(intermediate))
            result = (result[current_path] 
                if current_path in result 
                else None)
            if not result and not allowEmpty: raise Exception(
                "Null item found for [{}] in [{}]".format(
                    current_path, 
                    type(self).__name__))
            if not result: return []
            intermediate = current_path
        if not isinstance(result, list): raise Exception(
            "No list found at the specified path")
        return list(result)



from .item import Item
from .album import Album

class AlbumList:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getAlbums(self) -> list[Album]:
        l : list = self.__item.getList(["albumList2", "album"])
        return list(map(lambda x : Album(x), l))


from .item import Item
from .similar_artist import SimilarArtist

class ArtistInfo:

    __dict_name : str = "artistInfo2"

    def __init__(self, data : dict):
        self.__item : Item = Item(data)
        select_item : self.__item
        if self.__item.isResponse() and ArtistInfo.__dict_name in data:
            self.__is_response = True
            self.__response = self.__item
            select_item = Item(data[ArtistInfo.__dict_name])
        else:
            self.__is_response = False
            self.__response = None
            select_item = self.__item
        self.__select_item : Item = select_item

    def getItem(self): return self.__select_item

    def getBiography(self) -> str:
        return self.__select_item.getByName("biography")

    def getMusicBrainzId(self) -> str:
        return self.__select_item.getByName("musicBrainzId")

    def getLastFmUrl(self) -> str:
        return self.__select_item.getByName("lastFmUrl")

    def getSmallImageUrl(self) -> str:
        return self.__select_item.getByName("smallImageUrl")

    def getMediumImageUrl(self) -> str:
        return self.__select_item.getByName("mediumImageUrl")

    def getLargeImageUrl(self) -> str:
        return self.__select_item.getByName("largeImageUrl")

    def getSimilarArtists(self) -> list[SimilarArtist]:
        return list(map(
            lambda x : SimilarArtist(x), 
            self.__select_item.getList(["similarArtist"])))



from .item import Item
from .artists_initial import ArtistsInitial

class Artists:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getArtistListInitials(self) -> list[ArtistsInitial]:
        return list(map(
            lambda x : ArtistsInitial(x), 
            self.__item.getList(["artists", "index"])))


from .item import Item
from .album import Album

class Artist:

    __dict_name : str = "artist"

    def __init__(self, data : dict):
        self.__item : Item = Item(data)
        select_item : self.__item
        if self.__item.isResponse() and Artist.__dict_name in data:
            self.__is_response = True
            self.__response = self.__item
            select_item = Item(data[Artist.__dict_name])
        else:
            self.__is_response = False
            self.__response = None
            select_item = self.__item
        self.__select_item : Item = select_item

    def getItem(self): return self.__select_item

    def getId(self) -> str:
        return self.__select_item.getId()

    def getName(self) -> str:
        return self.__select_item.getByName("name")

    def getAlbumCount(self) -> str:
        return self.__select_item.getByName("albumCount")

    def getArtistImageUrl(self) -> str:
        return self.__select_item.getByName("artistImageUrl")

    def getAlbumList(self) -> list:
        albumList : list = self.__select_item.getList(["album"])
        return list(map(lambda x : Album(x), albumList))

    def getStarred(self) -> str:        
        return self.__select_item.getByName("starred")



class ArtistCover:

    def __init__(self, artist_id : str, artist_art_url : str, album_id : str = None):
        self._artist_id : str = artist_id
        self._artist_art_url : str = artist_art_url
        self._album_id : str = album_id

    def getArtistId(self) -> str:
        return self._artist_id

    def getArtistArtUrl(self) -> str:
        return self._artist_art_url
    
    def getAlbumId(self) -> str:
        return self._album_id



import libsonic

import urllib.parse

from .item import Item
from .artist import Artist
from .artists import Artists
from .album_list import AlbumList
from .album import Album
from .song import Song
from .random_songs import RandomSongs
from .top_songs import TopSongs
from .artist_info import ArtistInfo
from .search_result import SearchResult
from .artist_cover import ArtistCover
from .genres import Genres
from .playlists import Playlists
from .playlist import Playlist
from .response import Response
from .list_type import ListType
from .internet_radio_stations import InternetRadioStations
from .similar_songs import SimilarSongs
from .starred import Starred

from .configuration import Configuration


class Connector:

    def __init__(self, configuration: Configuration):
        self.__configuration = configuration

    def ping(self) -> bool:
        return self.__connect().ping()

    def getIndexes(
            self,
            musicFolderId=None,
            ifModifiedSince=0):
        return self.__connect().getIndexes(
            musicFolderId=musicFolderId,
            ifModifiedSince=ifModifiedSince)

    def getArtists(self) -> Response[Artists]:
        data: dict = self.__connect().getArtists()
        return Response(data, Artists(data) if data else None)

    def getGenres(self) -> Response[Genres]:
        data: dict = self.__connect().getGenres()
        return Response(data, Genres(data) if data else None)

    def getCoverArtForGenre(
            self,
            genre: str,
            maxSongs: int = 20) -> str:
        songs: dict = self.__connect().getSongsByGenre(
            genre,
            maxSongs)
        if not songs:
            raise Exception("No songs for genre [{}]".format(genre))
        songList = Item(songs).getList(["songsByGenre", "song"])
        if not songList:
            return None
        current_art: str = ""
        for current in songList:
            current_art = current["coverArt"] if "coverArt" in current else None
            if current_art:
                return current_art

    def getTopSongs(
            self,
            artist: str,
            count: int = 50) -> Response[TopSongs]:
        data: dict = self.__connect().getTopSongs(
            artist=artist,
            count=count)
        return Response(data, TopSongs(data) if data else None)

    def getArtistInfo(
            self,
            aid,
            count=20,
            includeNotPresent=False) -> Response[ArtistInfo]:
        data: dict = self.__connect().getArtistInfo2(
            aid=aid,
            count=count,
            includeNotPresent=includeNotPresent)
        return Response(data, ArtistInfo(data) if data else None)

    def getRandomSongs(
            self,
            size: int = 10,
            genre: any = None,
            fromYear=None,
            toYear=None,
            musicFolderId=None) -> Response[RandomSongs]:
        data: dict = self.__connect().getRandomSongs(
            size=size,
            genre=genre,
            fromYear=fromYear,
            toYear=toYear,
            musicFolderId=musicFolderId)
        return Response(data, RandomSongs(data) if data else None)

    def getAlbumList(
            self,
            ltype: ListType = None,
            size=10,
            offset=0,
            fromYear=None,
            toYear=None,
            genre=None,
            musicFolderId=None) -> Response[AlbumList]:
        data: dict = self.__connect().getAlbumList2(
            ltype=ltype.getArgValue() if ltype else None,
            size=size,
            offset=offset,
            fromYear=fromYear,
            toYear=toYear,
            genre=genre)
        return Response(data, AlbumList(data) if data else None)

    def getAlbum(
            self,
            albumId: str) -> Response[Album]:
        data: dict = self.__connect().getAlbum(albumId)
        return Response(data, Album(data) if data else None)

    def getSong(
            self,
            song_id: str) -> Response[Song]:
        data: dict = self.__connect().getSong(song_id)
        return Response(data, Song(data) if data else None)

    def getNewestAlbumList(
            self,
            size=10,
            offset=0,
            fromYear=None,
            toYear=None,
            genre=None,
            musicFolderId=None) -> Response[AlbumList]:
        return self.getAlbumList(
            ltype=ListType.NEWEST,
            size=size,
            offset=offset,
            fromYear=fromYear,
            toYear=toYear,
            genre=genre,
            musicFolderId=musicFolderId)

    def getRandomAlbumList(
            self,
            size=10,
            offset=0,
            fromYear=None,
            toYear=None,
            genre=None,
            musicFolderId=None) -> Response[AlbumList]:
        return self.getAlbumList(
            ltype=ListType.RANDOM,
            size=size,
            offset=offset,
            fromYear=fromYear,
            toYear=toYear,
            genre=genre,
            musicFolderId=musicFolderId)

    def getArtist(
            self,
            artist_id: str) -> Response[Artist]:
        data: dict = self.__connect().getArtist(artist_id)
        return Response(data, Artist(data) if data else None)

    def getPlaylists(self) -> Response[Playlists]:
        data: dict = self.__connect().getPlaylists()
        return Response(data, Playlists(data) if data else None)

    def getPlaylist(self, playlist_id: str) -> Response[Playlist]:
        data: dict = self.__connect().getPlaylist(playlist_id)
        return Response(data, Playlist(data) if data else None)

    def getCoverByArtistId(self, artist_id: str) -> ArtistCover:
        artist_response: Response[Artist] = self.getArtist(artist_id)
        if not artist_response and not artist_response.getObj():
            return None
        return self.getCoverByArtist(artist_response.getObj())

    def getCoverByArtist(self, artist: Artist) -> ArtistCover:
        artist_image_url: str = artist.getArtistImageUrl()
        album_list: list[Album] = artist.getAlbumList()
        selected_album: Album
        for selected_album in album_list:
            select_album_id = selected_album.getId()
            select_album_cover_art = selected_album.getCoverArt()
            if select_album_id and select_album_cover_art:
                return ArtistCover(
                    artist_id=artist.getId(),
                    album_id=select_album_id,
                    artist_art_url=artist_image_url)
        # no albums maybe?
        return ArtistCover(
            artist_id=artist.getId(),
            artist_art_url=artist_image_url)

    def search(
            self,
            query: str,
            artistCount: int = 20,
            artistOffset: int = 0,
            albumCount: int = 20,
            albumOffset: int = 0,
            songCount: int = 20,
            songOffset: int = 0,
            musicFolderId=None) -> SearchResult:
        return SearchResult(self.__connect().search3(
            query=query,
            artistCount=artistCount,
            artistOffset=artistOffset,
            albumCount=albumCount,
            albumOffset=albumOffset,
            songCount=songCount,
            songOffset=songOffset,
            musicFolderId=musicFolderId))

    def getStarred(self) -> Response[Starred]:
        data: dict = self.__connect().getStarred2()
        return Response(data, Starred(data) if data else None)

    def getInternetRadioStations(self) -> Response[InternetRadioStations]:
        data: dict = self.__connect().getInternetRadioStations()
        return Response(data, InternetRadioStations(data) if data else None)

    def getSimilarSongs(self, iid, count: int = 50) -> Response[SimilarSongs]:
        data: dict = self.__connect().getSimilarSongs2(iid=iid, count=count)
        return Response(data, SimilarSongs(data) if data else None)

    def buildSongUrlBySong(self, song: Song, format: str = None, max_bitrate: int = None) -> str:
        url_dict: dict[str, str] = dict()
        url_dict["id"] = song.getId()
        if format:
            url_dict["format"] = format
        if max_bitrate:
            url_dict["maxBitRate"] = max_bitrate
        return self.__buildUrl(
            verb="stream",
            url_dict=url_dict)

    def buildSongUrl(self, song_id: str, format: str = None, max_bitrate: int = None) -> str:
        song_res: Response[Song] = self.getSong(song_id)
        if song_res:
            return self.buildSongUrlBySong(
                song=song_res.getObj(),
                format=format,
                max_bitrate=max_bitrate)

    def scrobble(self, song_id: str, submission: bool = True, listenTime: int = None) -> dict:
        return self.__connect().scrobble(
            sid=song_id,
            submission=submission,
            listenTime=listenTime)

    def buildCoverArtUrl(self, item_id: str) -> str:
        return self.__buildUrl(
            verb="getCoverArt",
            url_dict={"id": item_id})

    def __addAuthParameters(
            self,
            url_dict: dict[str, str] = None) -> dict[str, str]:
        connection = self.__connect()
        qdict = connection._getBaseQdict()
        url_dict["u"] = self.__configuration.getUserName()
        if "s" in qdict:
            url_dict["s"] = qdict["s"]  # salt
        if "t" in qdict:
            url_dict["t"] = qdict["t"]  # token
        if "p" in qdict:
            url_dict["p"] = qdict["p"]  # password
        url_dict["c"] = self.__configuration.getAppName()
        url_dict["v"] = self.__configuration.getApiVersion()
        return url_dict

    def __buildUrl(self, verb, url_dict: dict[str, str] = None) -> str:
        url: str = "{}/rest/{}".format(self.__createBaseUrlWithPort(), verb)
        local_url_dict: dict[str, str] = (self.__addAuthParameters(url_dict)
                                          if url_dict
                                          else dict())
        local_url_dict["f"] = "json"
        params: str = urllib.parse.urlencode(local_url_dict, doseq=True)
        return f"{url}?{params}"


    def __createBaseUrlWithPort(self):
        baseUrl = self.__configuration.getBaseUrl()
        port = self.__configuration.getPort()
        url = baseUrl
        if ((baseUrl and baseUrl.startswith("https://") and port != 443) or
           (baseUrl and baseUrl.startswith("http://") and port != 80)):
            url = "{}:{}".format(baseUrl, port)
        return url

    def download(self, song_id: str):
        return self.__connect().download(song_id)

    def __connect(self):
        return libsonic.Connection(
            baseUrl=self.__configuration.getBaseUrl(),
            username=self.__configuration.getUserName(),
            password=self.__configuration.getPassword(),
            port=int(str(self.__configuration.getPort())),
            legacyAuth=self.__configuration.getLegacyAuth(),
            appName=str(self.__configuration.getAppName()),
            apiVersion=str(self.__configuration.getApiVersion()))



from .item import Item
from .song import Song
from .multi_value import get_multi

class Album:

    __key_title : str = "title"
    __key_name : str = "name"

    __dict_name : str = "album"

    def __init__(self, data : dict):
        self.__item: Item = Item(data)
        select_item = self.__item
        if self.__item.isResponse() and Album.__dict_name in data:
            self.__is_response = True
            self.__response = self.__item
            select_item = Item(data[Album.__dict_name])
        else:
            self.__is_response = False
            self.__response = None
            select_item = self.__item
        self.__select_item: Item = select_item

    def getItem(self): return self.__select_item

    def getId(self) -> str:
        return self.__select_item.getByName("id")

    def getArtist(self) -> str:
        return self.__select_item.getByName("artist")

    def getArtistId(self) -> str:
        return self.__select_item.getByName("artistId")

    def getCoverArt(self) -> str:
        return self.__select_item.getByName("coverArt")

    def getTitle(self) -> str:
        title: str = self.__select_item.getByName(Album.__key_title)
        if not title:
            title: str = self.__select_item.getByName(Album.__key_name)
        return title

    def getGenre(self) -> str:
        genre_list: list[str] = self.__get_genres()
        return genre_list[0] if genre_list and len(genre_list) > 0 else None

    def getGenres(self) -> list[str]:
        return self.__get_genres()

    def __get_genres(self) -> list[str]:
        return get_multi(self.__select_item, "genre", "genres", "name")

    def getYear(self) -> int:
        return self.__select_item.getByName("year")

    def getOriginalReleaseDate(self) -> str:
        ord_dict: dict[str, any] = self.__select_item.getByName("originalReleaseDate")
        if ord_dict is None:
            return None
        # split and return
        y: int = ord_dict["year"] if "year" in ord_dict else None
        if y is None:
            return None
        m: int = ord_dict["month"] if "month" in ord_dict else None
        d: int = ord_dict["day"] if "day" in ord_dict else None
        if m is None or d is None:
            return y
        # combine
        return f"{y:04}-{m:02}-{d:02}"

    def getOriginalReleaseYear(self) -> str:
        ord_v: str = self.getOriginalReleaseDate()
        if not ord_v or len(ord_v) < 4:
            return None
        return ord_v[0:4]

    def getOriginalYearWithYear(self) -> str:
        year: int = self.getYear()
        if not year:
            return None
        original_year: str = self.getOriginalReleaseYear()
        if not original_year or int(original_year) == year:
            return str(year)
        return f"{original_year} [{year}]"

    def getDuration(self) -> int:
        return self.__select_item.getByName("duration")

    def getSongCount(self) -> int:
        return self.__select_item.getByName("songCount")

    def getStarred(self) -> str:        
        return self.__select_item.getByName("starred")

    def getSongs(self) -> list[Song]:
        return list(map(
            lambda x : Song(x), 
            self.__select_item.getList(["song"])))



from .item import Item

class PlaylistEntry:

    def __init__(self, data : dict):
        self.__item : Item = Item(data)

    def getItem(self): return self.__item

    def getId(self) -> str:
        return self.__item.getByName("id")

    def getParent(self) -> str:
        return self.__item.getByName("parent")
    
    def getTitle(self) -> str:
        return self.__item.getByName("title")

    def getAlbum(self) -> str:
        return self.__item.getByName("album")

    def getArtist(self) -> str:
        return self.__item.getByName("artist")

    def getIsDir(self) -> str:
        return self.__item.getByName("isDir")

    def getCoverArt(self) -> str:
        return self.__item.getByName("coverArt")

    def getCreated(self) -> str:
        return self.__item.getByName("created")

    def getDuration(self) -> str:
        return self.__item.getByName("duration")

    def getBitRate(self) -> str:
        return self.__item.getByName("bitRate")
    
    def getTrack(self) -> str:
        return self.__item.getByName("track")

    def getYear(self) -> str:
        return self.__item.getByName("year")

    def getSize(self) -> str:
        return self.__item.getByName("size")

    def getSuffix(self) -> str:
        return self.__item.getByName("suffix")
    
    def getContentType(self) -> str:
        return self.__item.getByName("contentType")

    def getIsVideo(self) -> str:
        return self.__item.getByName("isVideo")

    def getPath(self) -> str:
        return self.__item.getByName("path")

    def getAlbumId(self) -> str:
        return self.__item.getByName("albumId")

    def getArtistId(self) -> str:
        return self.__item.getByName("artistId")

    def getType(self) -> str:
        return self.__item.getByName("type")



from .configuration import ConfigurationInterface

import libsonic

import os

class DefaultConfiguration(ConfigurationInterface):

    def __getParameter(self, name : str, default : str = None) -> str:
        return os.getenv(name, default)
    
    def getBaseUrl(self) -> str:
        return self.__getParameter("SUBSONIC_SERVER_URL")
    
    def getPort(self) -> str:
        return self.__getParameter("SUBSONIC_SERVER_PORT")
    
    def getUserName(self) -> str:
        return self.__getParameter("SUBSONIC_USERNAME")
    
    def getPassword(self) -> str:
        return self.__getParameter("SUBSONIC_PASSWORD")
    
    def getLegacyAuth(self) -> bool:
        legacy_auth_enabled_str : str = self.__getParameter("SUBSONIC_LEGACYAUTH", "false")
        if not legacy_auth_enabled_str.lower() in ['true', 'false']: 
            raise Exception(f"Invalid value for SUBSONIC_LEGACYAUTH [{legacy_auth_enabled_str}]")
        return legacy_auth_enabled_str == "true"

    def getApiVersion(self) -> str:
        return self.__getParameter("SUBSONIC_API_VERSION", libsonic.API_VERSION)

    def getAppName(self) -> str:
        return self.__getParameter("SUBSONIC_APP_NAME", "subsonic-connector")
    

    
